"""
This file contains the default
declaration of the keywords
used to declare/manipulate a mesh in Python
----
Luc Laurent - luc.laurent@lecnam.net -- 2021
"""

DFLT_MESH = 'connectivity'
DFLT_TYPE_ELEM = 'type'
DFLT_PHYS_GRP = 'physgrp'
DFLT_FIELD_DATA = 'data'
DFLT_FIELD_TYPE = 'type'
DFLT_FIELD_TYPE_NODAL = 'nodal'
DFLT_FIELD_TYPE_ELEMENT = 'elemental'
DFLT_FIELD_TYPE_NODAL_SCALAR = 'nodal_scalar'
DFLT_FIELD_TYPE_ELEMENT_SCALAR = 'elemental_scalar'
DFLT_FIELD_DIM = 'dim'
DFLT_FIELD_NAME = 'name'
DFLT_FIELD_STEPS = 'steps'
DFLT_FIELD_NBSTEPS = 'nbsteps'
DFLT_FIELD_NUMENTITIES = 'numentities'
